# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import importlib.resources
import os

import PIL.Image

from torchvision.transforms.functional import center_crop, to_tensor

from ...utils.rc import load_rc
from ..datamodule import CachingDataModule
from ..image_utils import remove_black_borders
from ..split import JSONDatabaseSplit
from ..typing import DatabaseSplit
from ..typing import RawDataLoader as _BaseRawDataLoader
from ..typing import Sample


class RawDataLoader(_BaseRawDataLoader):
    """A specialized raw-data-loader for the HIV-TB dataset."""

    datadir: str
    """This variable contains the base directory where the database raw data is
    stored."""

    def __init__(self):
        self.datadir = load_rc().get(
            "datadir.hivtb", os.path.realpath(os.curdir)
        )

    def sample(self, sample: tuple[str, int]) -> Sample:
        """Loads a single image sample from the disk.

        Parameters
        ----------
        sample
            A tuple containing the path suffix, within the dataset root folder,
            where to find the image to be loaded, and an integer, representing the
            sample label.


        Returns
        -------
            The sample representation
        """
        image = PIL.Image.open(os.path.join(self.datadir, sample[0])).convert(
            "L"
        )
        image = remove_black_borders(image)
        tensor = to_tensor(image)
        tensor = center_crop(tensor, min(*tensor.shape[1:]))

        # use the code below to view generated images
        # from torchvision.transforms.functional import to_pil_image
        # to_pil_image(tensor).show()
        # __import__("pdb").set_trace()

        return tensor, dict(label=sample[1], name=sample[0])  # type: ignore[arg-type]

    def label(self, sample: tuple[str, int]) -> int:
        """Loads a single image sample label from the disk.

        Parameters
        ----------
        sample
            A tuple containing the path suffix, within the dataset root folder,
            where to find the image to be loaded, and an integer, representing the
            sample label.


        Returns
        -------
            The integer label associated with the sample
        """
        return sample[1]


def make_split(basename: str) -> DatabaseSplit:
    """Returns a database split for the HIV-TB database."""

    return JSONDatabaseSplit(
        importlib.resources.files(__name__.rsplit(".", 1)[0]).joinpath(basename)
    )


class DataModule(CachingDataModule):
    """HIV-TB dataset for computer-aided diagnosis (only BMP files)

    * Database reference: [HIV-TB-2019]_
    * Original resolution, varying with most images being 2048 x 2500 pixels
      or 2500 x 2048 pixels, but not all.

    Data specifications:

    * Raw data input (on disk):

      * BMP (BMP3) and JPEG grayscale images encoded as 8-bit RGB, with
        varying resolution

    * Output image:

      * Transforms:

        * Load raw BMP or JPEG with :py:mod:`PIL`
        * Remove black borders
        * Convert to torch tensor
        * Torch center cropping to get square image

    * Final specifications

      * Grayscale, encoded as a single plane tensor, 32-bit floats,
        square at 2048 x 2048 pixels
      * Labels: 0 (healthy), 1 (active tuberculosis)
    """

    def __init__(self, split_filename: str):
        super().__init__(
            database_split=make_split(split_filename),
            raw_data_loader=RawDataLoader(),
        )
