# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import importlib.resources

from ..datamodule import CachingDataModule
from ..shenzhen.datamodule import RawDataLoader
from ..split import JSONDatabaseSplit
from ..typing import DatabaseSplit


def make_split(basename: str) -> DatabaseSplit:
    """Returns a database split for the Indian database."""

    return JSONDatabaseSplit(
        importlib.resources.files(__name__.rsplit(".", 1)[0]).joinpath(basename)
    )


class DataModule(CachingDataModule):
    """Indian collection dataset for computer-aided diagnosis.

    The Indian collection database has been established to foster research in
    computer-aided diagnosis of pulmonary diseases with a special focus on
    pulmonary tuberculosis (TB).  This database is also known as the "Database
    A/Database B" database.

    * Original images PNG, 8-bit grayscale, 1024 x 1024 pixels
    * Split reference: [INDIAN-2013]_ with 20% of train set for the validation set

    Data specifications:

    * Raw data input (on disk):

      * PNG RGB 8-bit depth images with "inverted" grayscale scale
      * Variable width and height

    * Output image:

      * Transforms:

        * Load raw PNG with :py:mod:`PIL`
        * Remove black borders
        * Convert to torch tensor
        * Torch center cropping to get square image

      * Final specifications:

        * Grayscale, encoded as a single plane tensor, 32-bit floats,
          square, with varying resolutions, depending on the input raw image
        * Labels: 0 (healthy), 1 (active tuberculosis)
    """

    def __init__(self, split_filename: str):
        super().__init__(
            database_split=make_split(split_filename),
            raw_data_loader=RawDataLoader(config_variable="datadir.indian"),
        )
