# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import importlib.resources
import os
import typing

import PIL.Image

from torchvision.transforms.functional import to_tensor

from ...utils.rc import load_rc
from ..datamodule import CachingDataModule
from ..split import JSONDatabaseSplit
from ..typing import DatabaseSplit
from ..typing import RawDataLoader as _BaseRawDataLoader
from ..typing import Sample

BoundingBoxAnnotation: typing.TypeAlias = tuple[int, int, int, int, int]
"""Location of TB radiological findings (latent or active)

Objects of this type carry bounding-box information of radiological findings on
the original 512x512 pixel images of TBX11k.  The radiological findings are
defined as such:

* 0/1: This sign is for latent TB (0), or active TB (1)
* xmin: horizontal position of bounding box upper-left corner, in pixels
* ymin: vertical position of bounding box upper-left corner, in pixels
* width: width of the bounding box, in pixels
* height: height of the bounding box, in pixels
"""

DatabaseSample: typing.TypeAlias = (
    tuple[str, int] | tuple[str, int, list[BoundingBoxAnnotation]]
)
"""Type of objects in our JSON representation for this database.

For healthy/sick (no TB)/latent TB cases, each sample is represented by
a filename, relative to the root of the installed database, followed by
the number 0 (negative class).

For active TB cases, each sample is represented by a filename, followed
by the number 1, and then by 1 or more 5-tuples with radiological
finding locations, as described above.
"""


class RawDataLoader(_BaseRawDataLoader):
    """A specialized raw-data-loader for the TBX11k dataset."""

    datadir: str
    """This variable contains the base directory where the database raw data is
    stored."""

    def __init__(self):
        self.datadir = load_rc().get(
            "datadir.tbx11k", os.path.realpath(os.curdir)
        )

    def sample(self, sample: DatabaseSample) -> Sample:
        """Loads a single image sample from the disk.

        Parameters
        ----------
        sample
            A tuple containing the path suffix, within the dataset root folder,
            where to find the image to be loaded, an integer, representing the
            sample label, and possible radiological findings represented by
            bounding boxes.


        Returns
        -------
            The sample representation
        """
        image = PIL.Image.open(os.path.join(self.datadir, sample[0]))
        tensor = to_tensor(image)

        # use the code below to view generated images
        # from torchvision.transforms.functional import to_pil_image
        # to_pil_image(tensor).show()
        # __import__("pdb").set_trace()

        return tensor, dict(label=sample[1], name=sample[0])  # type: ignore[arg-type]

    def label(self, sample: DatabaseSample) -> int:
        """Loads a single image sample label from the disk.

        Parameters
        ----------
        sample
            A tuple containing the path suffix, within the dataset root folder,
            where to find the image to be loaded, an integer, representing the
            sample label, and possible radiological findings represented by
            bounding boxes.


        Returns
        -------
            The integer label associated with the sample
        """
        return sample[1]

    def bbox_annotations(
        self, sample: DatabaseSample
    ) -> list[BoundingBoxAnnotation]:
        """Loads a single image sample label from the disk.

        Parameters
        ----------
        sample
            A tuple containing the path suffix, within the dataset root folder,
            where to find the image to be loaded, an integer, representing the
            sample label, and possible radiological findings represented by
            bounding boxes.


        Returns
        -------
            Bounding box annotations, if any available with the sample.
        """
        return sample[2] if len(sample) > 2 else []  # type: ignore


def make_split(basename: str) -> DatabaseSplit:
    """Returns a database split for the Montgomery database."""

    return JSONDatabaseSplit(
        importlib.resources.files(__name__.rsplit(".", 1)[0]).joinpath(basename)
    )


class DataModule(CachingDataModule):
    """TBX11k dataset for TB detection.

    * Database reference: [TBX11K-2020]_
    * The original dataset contains samples of healthy, sick (no TB), active
      and latent TB cases.  There is a total of 11702 samples in the database.
      Healthy and sick individuals are kept in separate folders.  Latent and
      active TB cases are merged in the same directory.  One must check the
      radiological annotations to understand if samples contain either, or both
      (latent and active TB) signs.
    * There is one case of patient (file ``imgs/tb/tb1199.png``), that is
      inside the ``tb`` folder, but contains no annotations.  This sample was
      **excluded** from our splits.
    * There are 30 cases of patients that have both active and latent TB
      radiological signs, over the entire database.  Those samples were also
      **excluded** from our splits:

      - imgs/tb/tb0135.png
      - imgs/tb/tb0142.png
      - imgs/tb/tb0154.png
      - imgs/tb/tb0167.png
      - imgs/tb/tb0190.png
      - imgs/tb/tb0246.png
      - imgs/tb/tb0255.png
      - imgs/tb/tb0279.png
      - imgs/tb/tb0284.png
      - imgs/tb/tb0350.png
      - imgs/tb/tb0378.png
      - imgs/tb/tb0392.png
      - imgs/tb/tb0395.png
      - imgs/tb/tb0501.png
      - imgs/tb/tb0506.png
      - imgs/tb/tb0526.png
      - imgs/tb/tb0543.png
      - imgs/tb/tb0639.png
      - imgs/tb/tb0640.png
      - imgs/tb/tb0667.png
      - imgs/tb/tb0676.png
      - imgs/tb/tb0713.png
      - imgs/tb/tb0786.png
      - imgs/tb/tb0870.png
      - imgs/tb/tb0875.png
      - imgs/tb/tb0945.png
      - imgs/tb/tb0949.png
      - imgs/tb/tb0968.png
      - imgs/tb/tb1104.png
      - imgs/tb/tb1143.png

    * Original train dataset samples:

      - Healthy: 3000
      - Sick (but no TB): 3000
      - Active TB only: 473
      - Latent TB only: 103
      - Both active and latent TB: 23
      - Unknown: 1
      - Total: 6600

    * Original validation dataset samples:

      - Healthy: 800
      - Sick (but no TB): 800
      - Latent TB only: 36
      - Active TB only: 157
      - Both active and latent TB: 7
      - Total: 1800

    * Original test dataset samples:

      - Unknown: 3302
      - Total: 3302

    * Because the test set does not have annotations, we generated train,
      validation and test datasets as such:

      - The original validation dataset becomes our test set.
      - The original train dataset is split into new train and validation
        datasets (validation ratio = 0.203 w.r.t. original train dataset size).
        The selection of samples is stratified (see comments through our split
        code, which is shipped alongside this file.)

    Data specifications:

    * Raw data input (on disk): PNG images 8 bits RGB, 512 x 512 pixels

    * Output image:

      * Transforms:

        - Load raw PNG with :py:mod:`PIL`

      * Final specifications:

        - RGB, encoded as a 3-plane tensor using 32-bit floats, square
          (512x512 pixels)
        - Labels: 0 (healthy, latent tb or sick but no tb depending on the
          protocol), 1 (active tuberculosis)
    """

    def __init__(self, split_filename: str):
        super().__init__(
            database_split=make_split(split_filename),
            raw_data_loader=RawDataLoader(),
        )
