# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import csv
import logging
import os
import shutil

import lightning.pytorch
import lightning.pytorch.callbacks
import lightning.pytorch.loggers
import torch.nn

from ..utils.resources import ResourceMonitor, cpu_constants, gpu_constants
from .callbacks import LoggingCallback
from .device import DeviceManager

logger = logging.getLogger(__name__)


def save_model_summary(
    output_folder: str,
    model: torch.nn.Module,
) -> tuple[lightning.pytorch.callbacks.ModelSummary, int]:
    """Saves a little summary of the model in a txt file.

    Parameters
    ----------

    output_folder
        output path

    model
        Network (e.g. driu, hed, unet)


    Returns
    -------
    summary
        The model summary in a text format

    total_parameters
        The number of parameters of the model
    """
    summary_path = os.path.join(output_folder, "model_summary.txt")
    logger.info(f"Saving model summary at {summary_path}...")
    with open(summary_path, "w") as f:
        summary = lightning.pytorch.utilities.model_summary.ModelSummary(
            model, max_depth=-1
        )
        f.write(str(summary))
    return (
        summary,
        lightning.pytorch.utilities.model_summary.ModelSummary(
            model
        ).total_parameters,
    )


def static_information_to_csv(
    static_logfile_name: str,
    device_type: str,
    model_size: int,
) -> None:
    """Saves the static information in a CSV file.

    Parameters
    ----------

    static_logfile_name
        The static file name which is a join between the output folder and
        "constant.csv"

    device_type
        The type of device we are using

    model_size
        The size of the model we will be training
    """
    if os.path.exists(static_logfile_name):
        backup = static_logfile_name + "~"
        if os.path.exists(backup):
            os.unlink(backup)
        shutil.move(static_logfile_name, backup)
    with open(static_logfile_name, "w", newline="") as f:
        logdata: dict[str, int | float | str] = {}
        logdata.update(cpu_constants())
        if device_type == "cuda":
            results = gpu_constants()
            if results is not None:
                logdata.update(results)
        logdata["model_size"] = model_size
        logwriter = csv.DictWriter(f, fieldnames=logdata.keys())
        logwriter.writeheader()
        logwriter.writerow(logdata)


def run(
    model: lightning.pytorch.LightningModule,
    datamodule: lightning.pytorch.LightningDataModule,
    checkpoint_period: int,
    device_manager: DeviceManager,
    max_epochs: int,
    output_folder: str,
    monitoring_interval: int | float,
    batch_chunk_count: int,
    checkpoint: str,
):
    """Fits a CNN model using supervised learning and save it to disk.

    This method supports periodic checkpointing and the output of a
    CSV-formatted log with the evolution of some figures during training.


    Parameters
    ----------

    model
        Neural network model (e.g. pasa).

    datamodule
        The lightning datamodule to use for training **and** validation

    checkpoint_period
        Save a checkpoint every ``n`` epochs.  If set to ``0`` (zero), then do
        not save intermediary checkpoints.

    device_manager
        An internal device representation, to be used for training and
        validation.  This representation can be converted into a pytorch device
        or a torch lightning accelerator setup.

    max_epochs
        The maximum number of epochs to train for.

    output_folder
        Directory in which the results will be saved.

    monitoring_interval
        Interval, in seconds (or fractions), through which we should monitor
        resources during training.

    batch_chunk_count
        If this number is different than 1, then each batch will be divided in
        this number of chunks.  Gradients will be accumulated to perform each
        mini-batch.   This is particularly interesting when one has limited RAM
        on the GPU, but would like to keep training with larger batches.  One
        exchanges for longer processing times in this case.

    checkpoint
    """

    os.makedirs(output_folder, exist_ok=True)

    # Save model summary
    _, no_of_parameters = save_model_summary(output_folder, model)

    from .loggers.custom_tensorboard_logger import CustomTensorboardLogger

    log_dir = "logs"
    tensorboard_logger = CustomTensorboardLogger(
        output_folder,
        log_dir,
    )
    logger.info(
        f"Monitor experiment with `tensorboard serve "
        f"--logdir={output_folder}/{log_dir}/`. "
        f"Then, open a browser on the printed address."
    )

    resource_monitor = ResourceMonitor(
        interval=monitoring_interval,
        has_gpu=device_manager.device_type == "cuda",
        main_pid=os.getpid(),
        logging_level=logging.ERROR,
    )

    checkpoint_callback = lightning.pytorch.callbacks.ModelCheckpoint(
        output_folder,
        "model_lowest_valid_loss",
        save_last=True,
        monitor="loss/validation",
        mode="min",
        save_on_train_epoch_end=True,
        every_n_epochs=checkpoint_period,
    )

    checkpoint_callback.CHECKPOINT_NAME_LAST = "model_final_epoch"

    # write static information to a CSV file
    static_logfile_name = os.path.join(output_folder, "constants.csv")
    static_information_to_csv(
        static_logfile_name,
        device_manager.device_type,
        no_of_parameters,
    )

    with resource_monitor:
        accelerator, devices = device_manager.lightning_accelerator()
        trainer = lightning.pytorch.Trainer(
            accelerator=accelerator,
            devices=devices,
            max_epochs=max_epochs,
            accumulate_grad_batches=batch_chunk_count,
            logger=tensorboard_logger,
            check_val_every_n_epoch=1,
            callbacks=[LoggingCallback(resource_monitor), checkpoint_callback],
        )

        _ = trainer.fit(model, datamodule, ckpt_path=checkpoint)
