# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
"""Import copy import os import shutil.

import numpy as np
import torch

from matplotlib.backends.backend_pdf import PdfPages
from sklearn import metrics
from torch.utils.data import ConcatDataset, DataLoader

from ..engine.predictor import run
from ..utils.plot import relevance_analysis_plot

logger = setup(__name__.split(".")[0], format="%(levelname)s: %(message)s")


# Relevance analysis using permutation feature importance
if relevance_analysis:
    if isinstance(v, ConcatDataset) or not isinstance(
        v._samples[0].data["data"], list
    ):
        logger.info(
            "Relevance analysis only possible with radiological signs as input. Cancelling..."
        )
        continue

    nb_features = len(v._samples[0].data["data"])

    if nb_features == 1:
        logger.info("Relevance analysis not possible with one feature")
    else:
        logger.info(f"Starting relevance analysis for subset '{k}'...")

        all_mse = []
        for f in range(nb_features):
            v_original = copy.deepcopy(v)

            # Randomly permute feature values from all samples
            v.random_permute(f)

            data_loader = DataLoader(
                dataset=v,
                batch_size=batch_size,
                shuffle=False,
                pin_memory=torch.cuda.is_available(),
            )

            predictions_with_mean = run(
                model,
                data_loader,
                k,
                accelerator,
                output_folder + "_temp",
            )

            # Compute MSE between original and new predictions
            all_mse.append(
                metrics.mean_squared_error(
                    np.array(predictions, dtype=object)[:, 1],
                    np.array(predictions_with_mean, dtype=object)[:, 1],
                )
            )

            # Back to original values
            v = v_original

        # Remove temporary folder
        shutil.rmtree(output_folder + "_temp", ignore_errors=True)

        filepath = os.path.join(output_folder, k + "_RA.pdf")
        logger.info(f"Creating and saving plot at {filepath}...")
        os.makedirs(os.path.dirname(filepath), exist_ok=True)
        pdf = PdfPages(filepath)
        pdf.savefig(
            relevance_analysis_plot(
                all_mse,
                title=k.capitalize() + " set relevance analysis",
            )
        )
        pdf.close()
"""
