# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

import os

import click
import matplotlib.pyplot as plt
import pandas

from clapper.click import ConfigCommand, ResourceOption, verbosity_option
from clapper.logging import setup
from matplotlib.ticker import MaxNLocator

logger = setup(__name__.split(".")[0], format="%(levelname)s: %(message)s")


def create_figures(df: pandas.DataFrame) -> list[plt.figure]:
    """Generates figures for each metric in the dataframe.

    Each row of the dataframe correspond to an epoch and each column to a metric.
    It is assumed that some metric names are of the form <metric>/<subset>.
    All subsets for a metric will be displayed on the same figure.


    Parameters
    ----------

    df:
        Pandas dataframe containing the data to plot.


    Returns
    -------

    figures:
        List of matplotlib figures, one per metric.
    """

    figures = []

    labels = sorted(df.columns)
    from collections import defaultdict

    # Dict of metric: subset. Subset can be None.
    metrics_groups = defaultdict(list)

    for label in labels:
        # Separate the name of the subset from the metric
        split_label = label.rsplit("/", 1)
        metric = split_label[0]
        subset = split_label[1] if len(split_label) > 1 else None
        metrics_groups[metric].append(subset)

    for metric, subsets in metrics_groups.items():
        figure = plt.figure()
        axes = figure.gca()

        for subset in subsets:
            if subset is None:
                axes.plot(
                    df["step"].values,
                    df[metric],
                    label=metric,
                )
            else:
                axes.plot(
                    df["step"].values,
                    df[metric + "/" + subset],
                    label=subset,
                )

        axes.xaxis.set_major_locator(MaxNLocator(integer=True))
        axes.set_title(metric)
        axes.set_xlabel("Epoch")

        axes.legend(loc="best")
        axes.grid(alpha=0.3)
        figure.set_layout_engine("tight")

        figures.append(figure)

    return figures


@click.command(
    entry_point_group="ptbench.config",
    cls=ConfigCommand,
    epilog="""Examples:

\b
    1. Analyzes a training log and produces various plots:

       .. code:: sh

          ptbench train-analysis -vv results/logs
""",
)
@click.argument(
    "logdir",
    type=click.Path(dir_okay=True, exists=True),
)
@click.option(
    "--output-pdf",
    "-o",
    help="Name of the output file to dump",
    required=True,
    show_default=True,
    default="trainlog.pdf",
)
@verbosity_option(logger=logger, cls=ResourceOption, expose_value=False)
def train_analysis(
    logdir: str,
    output_pdf: str,
    **_,
) -> None:
    """Creates a plot for each metric in the training logs and saves them in a
    pdf file."""

    from matplotlib.backends.backend_pdf import PdfPages

    from ..utils.tensorboard import get_scalars

    data = get_scalars(logdir)

    # makes sure the directory to save the output PDF is there
    dirname = os.path.dirname(os.path.realpath(output_pdf))
    if not os.path.exists(dirname):
        os.makedirs(dirname)

    with PdfPages(output_pdf) as pdf:
        for figure in create_figures(data):
            pdf.savefig(figure)
            plt.close(figure)
