# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
import importlib.metadata
import logging
import os
import pathlib
import sys
import time

logger = logging.getLogger(__name__)


def save_sh_command(output_folder: str | pathlib.Path) -> None:
    """Records command-line to reproduce this script.

    This function can record the current command-line used to call the script
    being run.  It creates an executable ``bash`` script setting up the current
    working directory and activating a conda environment, if needed.  It
    records further information on the date and time the script was run and the
    version of the package.


    Parameters
    ----------

    output_folder : str
        Path leading to the directory where the commands to reproduce the current
        run will be recorded. A subdirectory will be created each time this function
        is called to match lightning's versioning convention for loggers.
    """

    if isinstance(output_folder, str):
        output_folder = pathlib.Path(output_folder)

    destfile = output_folder / "command.sh"

    logger.info(f"Writing command-line for reproduction at '{destfile}'...")
    os.makedirs(output_folder, exist_ok=True)

    package = __name__.split(".", 1)[0]
    version = importlib.metadata.version(package)

    with destfile.open("w") as f:
        f.write("#!/usr/bin/env sh\n")
        f.write(f"# date: {time.asctime()}\n")
        f.write(f"# version: {version} ({package})\n")
        f.write(f"# platform: {sys.platform}\n")
        f.write("\n")
        args = []
        for k in sys.argv:
            if " " in k:
                args.append(f'"{k}"')
            else:
                args.append(k)
        if os.environ.get("CONDA_DEFAULT_ENV") is not None:
            f.write(f"# conda activate {os.environ['CONDA_DEFAULT_ENV']}\n")
        f.write(f"# cd {os.path.realpath(os.curdir)}\n")
        f.write(" ".join(args) + "\n")
    os.chmod(destfile, 0o755)
